<?php

namespace Drupal\ai_eca_agents\Plugin\AiAgentValidation;

use Drupal\ai_eca_agents\Services\ModelMapper\ModelMapperInterface;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\ai\OperationType\Chat\ChatMessage;
use Drupal\ai\Service\PromptJsonDecoder\PromptJsonDecoderInterface;
use Drupal\ai_agents\Attribute\AiAgentValidation;
use Drupal\ai_agents\Exception\AgentRetryableValidationException;
use Drupal\ai_agents\PluginBase\AiAgentValidationPluginBase;
use Drupal\ai_agents\PluginInterfaces\AiAgentValidationInterface;
use Drupal\Core\TypedData\Exception\MissingDataException;
use Illuminate\Support\Arr;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * The validation of the ECA agent.
 */
#[AiAgentValidation(
  id: 'eca_validation',
  label: new TranslatableMarkup('ECA Validation'),
)]
class EcaValidation extends AiAgentValidationPluginBase implements ContainerFactoryPluginInterface {

  /**
   * The JSON-decoder of the prompt.
   *
   * @var \Drupal\ai\Service\PromptJsonDecoder\PromptJsonDecoderInterface
   */
  protected PromptJsonDecoderInterface $promptJsonDecoder;

  /**
   * The model mapper.
   *
   * @var \Drupal\ai_eca_agents\Services\ModelMapper\ModelMapperInterface
   */
  protected ModelMapperInterface $modelMapper;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): AiAgentValidationInterface {
    $instance = new static($configuration, $plugin_id, $plugin_definition);
    $instance->promptJsonDecoder = $container->get('ai.prompt_json_decode');
    $instance->modelMapper = $container->get('ai_eca_agents.services.model_mapper');

    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultValidation(mixed $data): bool {
    $data = $this->decodeData($data);
    if (empty($data)) {
      throw new AgentRetryableValidationException(
        'The LLM response failed validation: could not decode from JSON.',
        0,
        NULL,
        'You MUST only provide a RFC8259 compliant JSON response.'
      );
    }

    if (!Arr::has($data, '0.message')) {
      throw new AgentRetryableValidationException(
        'The LLM response failed validation: feedback was not provided.',
        0,
        NULL,
        'You MUST provide a summary of your action or give feedback why you failed.'
      );
    }

    // Stop validation if the LLM failed to return a response.
    if (Arr::get($data, '0.action', 'fail')) {
      return TRUE;
    }

    // Validate the response against ECA-model schema.
    try {
      $this->modelMapper->fromPayload(Json::decode(Arr::get($data, '0.model')));
    }
    catch (MissingDataException $e) {
      throw new AgentRetryableValidationException(
        'The LLM response failed validation: the model definition was violated.',
        0,
        NULL,
        sprintf('Fix the following violations: %s.', $e->getMessage())
      );
    }

    return TRUE;
  }

  /**
   * Decodes the source data.
   *
   * @param mixed $source
   *   The source.
   *
   * @return array|null
   *   Returns the decoded data or NULL.
   */
  protected function decodeData(mixed $source): ?array {
    $text = NULL;

    switch (TRUE) {
      case $source instanceof ChatMessage:
        return $this->promptJsonDecoder->decode($source);

      case is_string($source):
        $text = $source;
        break;
    }

    return Json::decode($text ?? '');
  }

}
